﻿using IFinalPDF;
using KCOM_FinalService.Common;
using KCOMDataModel.Common;
using KCOMDataModel.DataModel;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.Remoting.Channels.Tcp;
using System.Threading;
using System.Threading.Tasks;

namespace KCOM_FinalService.Remoting
{
    public class RemFinalPDFStation : InterfaceFinalPDF, IDisposable
    {
        #region Properties
        TcpChannel _Channel = null;
        ServiceEventLog _Log = null;
        Thread _FinalPDFThread = null;
        string _NoticeMessageBase = null;
        string _NoticeTitleBase = null;
        string _MyIP = null;
        int _TypeName = 1;
        readonly KCOM_FinalService_Console.Properties.Settings _SettingPlace = null;

        //Queue<FINAL_PDF> _WorkItem = new Queue<FINAL_PDF>();
        ConcurrentQueue<FINAL_PDF> WorkingQueue = null;
        List<FINAL_PDF> WaitThread = null;
        List<FINAL_PDF> WorkingThread = null;

        private Type _UploadFinalType { get; set; }   /// upload final class type
            #endregion

        public RemFinalPDFStation()
        {
            System.Diagnostics.EventLog _eventLog = new System.Diagnostics.EventLog();
           

            /// get legacy interface for upload final pdf file
            string UploadFinalDllPath = System.IO.Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "UploadFinal.dll");
           
            if (File.Exists(UploadFinalDllPath))
            {
                Assembly asm = Assembly.LoadFrom(UploadFinalDllPath);
                this._UploadFinalType = asm.GetType("UploadFinal.UploadFinal");
            }
            /// up to here

            Init();
        }

        public RemFinalPDFStation(int MultipleFinalCount)
        {
            System.Diagnostics.EventLog _eventlog = new System.Diagnostics.EventLog();
         

            /// get legacy interface for upload final pdf file
            string UploadFinalDllPath = System.IO.Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "UploadFinal.dll");
         
            if (File.Exists(UploadFinalDllPath))
            {
                Assembly asm = Assembly.LoadFrom(UploadFinalDllPath);
                this._UploadFinalType = asm.GetType("UploadFinal.UploadFinal");
            }
            /// up to here

            Init();
        }

        public RemFinalPDFStation(ServiceEventLog eventLog)
        {
            _Log = eventLog;
            Init();
        }

      

        public void Init()
        {
            try
            {
                //_MyIP = System.Net.Dns.Resolve(System.Net.Dns.GetHostName()).AddressList[0].ToString();
                using (KCOMDataModel.DataModel.KCOMEntities _systemEntity = new KCOMDataModel.DataModel.KCOMEntities(KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString()))
                {
#if DEBUG
                    _TypeName = 1; //FINAL은 현재 타입이 1임
#else
                    _TypeName = 1;
                    ///TODO: check below code
                    ///_TypeName = _systemEntity.SERVICE_PROPERTIES.Where(data => data.SERVICE_ADDRESS == _MyIP).FirstOrDefault().SERVICET_TYPE;
#endif
                }
                _Log.Write("ConnectString :", KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString());
                _Channel = new TcpChannel(KCOM_FinalService_Console.Properties.Settings.Default.RemotingPort);
                _Log.Write("Port", KCOM_FinalService_Console.Properties.Settings.Default.RemotingPort.ToString());
                System.Runtime.Remoting.Channels.ChannelServices.RegisterChannel(_Channel, false);
                System.Runtime.Remoting.RemotingConfiguration.RegisterWellKnownServiceType(typeof(RemFinalPDFObject), "remFinalPDF", System.Runtime.Remoting.WellKnownObjectMode.Singleton);
                WorkingQueue = new ConcurrentQueue<FINAL_PDF>();                
                WorkingThread = new List<FINAL_PDF>();
                WaitThread = new List<FINAL_PDF>();
                if(WorkingQueue !=null)
                    _Log.Write("WorkingQueue : ", WorkingQueue.GetType().Name);
                else
                    _Log.Write("WorkingQueue : ", "null");
            }
            catch (Exception EX)
            {
                if (_Log != null)
                {
                
                }
            }
            IFinalPDF.Cache.Attach(this);
        }

        private void SetFinalState(string finalID, FinalStatus status)
        {
            _Log.Write("finalID : ", finalID);

            using (KCOMDataModel.DataModel.KCOMEntities _entity = new KCOMDataModel.DataModel.KCOMEntities(KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString()))
            {
                var finalList = _entity.FINAL_PDF.Where(final => final.ID == finalID);

                if (finalList.Count() > 0)
                {
                    if (status == FinalStatus.Create)
                    {
                        finalList.First().START_DATETIME = DateTime.Now;
                    }
                    if (status == FinalStatus.Success)
                    {
                        finalList.First().END_DATETIME = DateTime.Now;
                    }

                    finalList.First().STATUS = (int)status;
                    _entity.SaveChanges();
                }
            };
        }

        public void Dispose()
        {
            if (_Channel != null)
            {
                System.Runtime.Remoting.Channels.ChannelServices.UnregisterChannel(_Channel);
            }
        }
        private void FinalPDFEnqueue(FINAL_PDF pdf)
        {
            _Log.Write("FinalPDFEnqueue : ", pdf.MARKUPINFO_ID);
            Task tEnq = Task.Factory.StartNew(() =>
            {
                WorkingQueue.Enqueue(pdf);
                Thread.Sleep(100);                
            });
            FinalPDFDequeueThreadStart();
        }        
        private void FinalPDFDequeueThreadStart()
        {
            _FinalPDFThread = new Thread(new ThreadStart(FinalPDFDequeueandProcessing));
            _FinalPDFThread.SetApartmentState(ApartmentState.STA);
            _FinalPDFThread.Start();
        }        
        
        private void FinalPDFDequeueandProcessing()
        {
            Task tDeq = Task.Factory.StartNew(() =>
            {
                _Log.Write("FinalService", "StartNew()");

                FINAL_PDF _item = new FINAL_PDF();


                try
                {
                    if (WorkingQueue == null)
                    {
                        WorkingQueue = new ConcurrentQueue<FINAL_PDF>();
                    }
                    _Log.Write("WorkingQueue", WorkingQueue.Count.ToString());
                    while (WorkingQueue.Count > 0)
                    {

                        if (WorkingQueue.TryDequeue(out _item))
                        {
                            _Log.Write("TryDequeue _item", _item.ID);

                            if (WorkingThread.Count() < KCOM_FinalService_Console.Properties.Settings.Default.MultipleFinalCount)
                            {
                                if (_item.STATUS == (int)FinalStatus.Insert)
                                {
                                    _item.STATUS = (int)(FinalStatus.Wait);
                                    using (KCOMDataModel.DataModel.KCOMEntities _entity = new KCOMDataModel.DataModel.KCOMEntities(KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString()))
                                    {
                                        var selectItem = _entity.FINAL_PDF.Where(data => data.ID == _item.ID).FirstOrDefault();

                                        selectItem.STATUS = (int)(FinalStatus.Wait);

                                        _entity.SaveChanges();
                                    }
                                }
                                var _identPathItem = WorkingThread.Where(item => item.DOCUMENT_ID == _item.DOCUMENT_ID);                                

                                if (_identPathItem.Count() > 0)
                                {
                                    WaitThread.Add(_item);
                                }
                                else
                                {
                                    MarkupToPDF.MarkupToPDF _markuptoPDF = new MarkupToPDF.MarkupToPDF();
                                    _markuptoPDF.FinalMakeError += new EventHandler<MarkupToPDF.MakeFinalErrorArgs>(_markuptoPDF_FinalMakeError);
                                    _markuptoPDF.EndFinal += new EventHandler<MarkupToPDF.EndFinalEventArgs>(_markuptoPDF_EndFinal);
                                    Thread _stathread = new Thread(new ParameterizedThreadStart(_markuptoPDF.MakeFinalPDF))
                                    {
                                        Name = _item.ID.ToString()
                                    };
                                    SetFinalState(_item.ID, IFinalPDF.FinalStatus.Create);
                                    //_stathread.Priority = ThreadPriority.Normal;
                                    _stathread.SetApartmentState(ApartmentState.STA);
                                    _stathread.Start(_item);
                                    WorkingThread.Add(_item);
                                }
                            }
                        }
                        Thread.Sleep(100);
                    }
                }
                catch (Exception ex)
                {
                    _Log.Write("FinalService", "FinalPDFDequeueandProcessing error:" + ex.ToString());
                }
                
            });
        }
        public FinalPDFResult Notify(string ProjectNo, string FinalID)
        {
            using (KCOMDataModel.DataModel.KCOMEntities _entity = new KCOMDataModel.DataModel.KCOMEntities(KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString()))
            {
                var finalpdf = _entity.FINAL_PDF.Where(final => final.ID == FinalID).FirstOrDefault();
                FinalPDFEnqueue(finalpdf);
                //FinalPDFProcessStart(finalList.ToList());
            }            
            ///TODO: why return FinalPDFResult?
            return new FinalPDFResult();
        }

        void _markuptoPDF_EndFinal(object sender, MarkupToPDF.EndFinalEventArgs e)
        {
            var _T = WorkingThread.Where(t => t == (sender as MarkupToPDF.MarkupToPDF).FinalItem);

            if (_T.Count() > 0)
            {
                FINAL_PDF _item = _T.First();

                SetFinalState(_item.ID, FinalStatus.PdfStamp);

                try
                {
                    string soapurl = CommonLib.Common.GetConfigString("UpLoadServiceUrl", "URL", "");
                    string finalpdfroot = CommonLib.Common.GetConfigString("FinalPDFRootUrl", "URL", "");
                    string filename = Path.GetFileName(e.FinalPDFPath);
                    string pdfpath = !string.IsNullOrEmpty(finalpdfroot) ? finalpdfroot + filename : filename;
                    if (!string.IsNullOrEmpty(soapurl) && (this._UploadFinalType != null))
                    {
                        //Legacy IF
                        MethodInfo UploadFinalPDFMethod = this._UploadFinalType.GetMethod("UploadFinalPDF", new Type[] { typeof(string), typeof(string), typeof(FINAL_PDF), typeof(string) });
                        if (null != UploadFinalPDFMethod)
                        {
                            object inst = Activator.CreateInstance(this._UploadFinalType, null);
                            object[] parameters = new object[] { pdfpath, e.OriginPDFName, e.FinalPDF, soapurl };
                            KeyValuePair<bool, string> result = (KeyValuePair<bool, string>)(UploadFinalPDFMethod.Invoke(inst, parameters));
                            ///KeyValuePair<bool, string> result = UploadFinal.UploadFinal.UploadFinalPDF(pdfpath, e.OriginPDFName, e.FinalPDF, soapurl);
                            if (result.Key)
                            {
                                SetFinalState(_item.ID, FinalStatus.Success);
                            }
                            else
                            {
                                SetFinalState(_item.ID, FinalStatus.Error);
                                _Log.Write("FinalService", "Upload error .." + filename + ",pdfname:" + e.OriginPDFName);
                            }
                        }
                        WorkingThread.Remove(_T.First());
                    }
                    else
                    {
                        ///TODO: 저장할 폴더 위치를 configuration으로 빼주세요

                        string savepath = CommonLib.Common.GetConfigString("DebugSavePath", "URL", "");

                        string saveFolder = String.Format(savepath, _item.PROJECT_NO, Convert.ToInt32(_item.DOCUMENT_ID) / 100, _item.DOCUMENT_ID);
                        if (!System.IO.Directory.Exists(saveFolder))
                        {
                            System.IO.Directory.CreateDirectory(saveFolder);
                        }

                        this._Log.Write("FinalService", String.Format("saveFolder : {0}", saveFolder) + DateTime.Now);
                        try
                        {
                            System.IO.File.Copy(e.FinalPDFPath, saveFolder + e.OriginPDFName, true);
                        }
                        catch (Exception)
                        {
                            _Log.Write("FinalService","e.FinalPDFPath:" + e.FinalPDFPath + ",saveFolder:" + saveFolder + ",e.OriginPDFName:" + e.OriginPDFName);
                        }

                        using (CIEntities _entity = new CIEntities(ConnectStringBuilder.ProjectCIConnectString(_item.PROJECT_NO).ToString()))
                        {
                            var item = _entity.DOCUMENT_ITEM.Where(d => d.DOCUMENT_ID == _item.DOCUMENT_ID).FirstOrDefault();
                            ///TODO: RESULT FILE 경로 위치를 configuration으로 빼주세요
                            string resultpath = CommonLib.Common.GetConfigString("DebugResultUrlPath", "URL", "");

                            item.RESULT_FILE = String.Format(resultpath, _item.PROJECT_NO, Convert.ToInt32(_item.DOCUMENT_ID) / 100, _item.DOCUMENT_ID, e.OriginPDFName);
                            sendReqLog("RESULT_FILE_PATH", item.RESULT_FILE);
                            _entity.SaveChanges();
                            SetFinalState(_item.ID, FinalStatus.Success);
                            WorkingThread.Remove(_T.First());
                        }
                    }
                }
                catch (Exception ex)
                {
                    SetFinalState(_item.ID, FinalStatus.Error);
                    _Log.Write("FinalService", "에러 .." + ex.Message);
                    WorkingThread.Remove(_T.First());
                }

                var waitpdf = WaitThread.Where(t => t.DOCUMENT_ID == _item.DOCUMENT_ID).FirstOrDefault();
                if(waitpdf != null)
                {
                    FinalPDFEnqueue(waitpdf);
                    WaitThread.Remove(waitpdf);
                }
                
            }
        }



        public static void sendReqLog(string name, string text)
        {
#if DEBUG
            try
            {
                string Msg = DateTime.Now.ToString("yyyy-MM-dd hh:mm:ss") + "\t" + name + "\t" + text + "\r\n";
                string Dir = Path.Combine(@"D:\finalpdf\error");

                if (Directory.Exists(Dir) == false)
                {
                    Directory.CreateDirectory(Dir);
                }

                int min = DateTime.Now.Minute / 1;

                string path = Dir + "/" + "Log_" + DateTime.Now.ToString("yyyy-MM-dd HH-") + min + ".txt";
                File.AppendAllText(path, Msg);
            }
            catch (Exception e)
            {
                //  Debug.WriteLine(e);
            }
#endif
        }


        private void SetError(string FinalID, string Msg)
        {
            FINAL_PDF instanceFinal = null;

            try
            {
                using (KCOMDataModel.DataModel.KCOMEntities _entity = new KCOMDataModel.DataModel.KCOMEntities(KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString()))
                {
                    instanceFinal = _entity.FINAL_PDF.Where(data => data.ID == FinalID).FirstOrDefault();

                    if (instanceFinal != null)
                    {
                        instanceFinal.STATUS = (int)FinalStatus.Error;
                        instanceFinal.EXCEPTION = DateTime.Now.ToShortDateString() + " " + Msg;
                        _entity.SaveChanges();
                    }
                }
            }
            catch (Exception ex)
            {
                _Log.Write("FinalService", ex.ToString());
            }

            /// 에러가 발생했을때 메일 발송
            //try //이메일 발송 부분 아직 보류
            //{

            //}
            //catch (Exception)
            //{

            //    throw;
            //}
            /// up to here
        }

        void _markuptoPDF_FinalMakeError(object sender, MarkupToPDF.MakeFinalErrorArgs e)
        {
            SetError(e.FinalID, e.Message);

            var _T = WorkingThread.Where(t => t == (sender as MarkupToPDF.MarkupToPDF).FinalItem);

            if (_T.Count() > 0)
            {
                FINAL_PDF _Item = _T.First();
                (sender as MarkupToPDF.MarkupToPDF).FinalMakeError -= new EventHandler<MarkupToPDF.MakeFinalErrorArgs>(_markuptoPDF_FinalMakeError);
                (sender as MarkupToPDF.MarkupToPDF).EndFinal -= new EventHandler<MarkupToPDF.EndFinalEventArgs>(_markuptoPDF_EndFinal);
                (sender as MarkupToPDF.MarkupToPDF).Dispose();
                WorkingThread.Remove(_Item);

                //Console.WriteLine(_Item.ID + "의 작업 중 에러가 발생했습니다");
            }
        }

        //private void CurrentCountUpdater()
        //{
        //    using (KCOMDataModel.DataModel.KCOMEntities _systemEntity = new KCOMDataModel.DataModel.KCOMEntities(KCOMDataModel.Common.ConnectStringBuilder.KCOMConnectionString().ToString()))
        //    {
        //        var items = _systemEntity.SERVICE_PROPERTIES.Where(data => data.SERVICET_TYPE == _TypeName).ToList();

        //        items.Where(data => data.Property == "PROCESS_COUNT").FirstOrDefault().Value = _WorkItems.Count.ToString();

        //        _system.SaveChanges();
        //    }
        //}
    }
}
